<?php

namespace Drupal\config_action_report\Controller;

use Drupal\Core\Config\Action\Attribute\ActionMethod;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Config\Entity\ConfigEntityTypeInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Symfony\Component\String\Inflector\EnglishInflector;

/**
 * Controller for Config Action methods report.
 */
class ConfigActionReportMethodsController extends ControllerBase implements ContainerInjectionInterface {

  /**
   * Displays an Config Action methods report.
   *
   * @return array
   *   A render array containing an Config Action methods report.
   *
   * @see \Drupal\Core\Config\Action\Plugin\ConfigAction\Deriver\EntityMethodDeriver::getDerivativeDefinitions
   */
  public function index(): array {
    // Header.
    $header = [
      'entity_type' => ['data' => $this->t('Entity type'), 'width' => '20%'],
      'label' => ['data' => $this->t('Label'), 'width' => '20%'],
      'method' => ['data' => $this->t('Action method'), 'width' => '20%'],
      'method_plural' => ['data' => $this->t('Action method plural'), 'width' => '20%'],
      'parameters' => ['data' => $this->t('Action parameters'), 'width' => '20%'],
    ];

    // Rows.
    $rows = [];

    // Get global config action methods.
    $configEntityClass = ConfigEntityBase::class;
    $reflectionClass = new \ReflectionClass($configEntityClass);
    foreach ($reflectionClass->getMethods(\ReflectionMethod::IS_PUBLIC) as $method) {
      $this->appendMethodToRows('*all*', $method, $rows);
    }

    foreach ($this->entityTypeManager()->getDefinitions() as $entity_type) {
      if (!$entity_type instanceof ConfigEntityTypeInterface) {
        continue;
      }

      $reflectionClass = new \ReflectionClass($entity_type->getClass());
      while ($reflectionClass && $reflectionClass->getName() !== $configEntityClass) {
        foreach ($reflectionClass->getMethods(\ReflectionMethod::IS_PUBLIC) as $method) {
          if ($method->getDeclaringClass()->getName() === $reflectionClass->getName()) {
            $this->appendMethodToRows($entity_type->id(), $method, $rows);
          }
        }
        $reflectionClass = $reflectionClass->getParentClass();
      }
    }

    ksort($rows);

    return [
      '#theme' => 'table',
      '#header' => $header,
      '#rows' => $rows,
    ];
  }

  /**
   * Appends method details to the rows array for rendering.
   *
   * @param string $entity_type
   *   The entity type ID or identifier.
   * @param \ReflectionMethod $method
   *   The ReflectionMethod object representing the method to analyze.
   * @param array &$rows
   *   The rows array, passed by reference, to which method details are added.
   */
  protected function appendMethodToRows(string $entity_type, \ReflectionMethod $method, array &$rows): void {
    $inflector = new EnglishInflector();
    foreach ($method->getAttributes(ActionMethod::class) as $attribute) {
      /** @var \Drupal\Core\Config\Action\Attribute\ActionMethod $attribute_instance */
      $attribute_instance = $attribute->newInstance();
      $method_name = $attribute_instance->name ?? $method->getName();
      $method_plural = '';
      if (!empty($attribute_instance->pluralize)) {
        $method_plural = (is_string($attribute_instance->pluralize))
          ? $attribute_instance->pluralize
          : $inflector->pluralize($method_name);
      }

      $parameters = [];
      foreach ($method->getParameters() as $parameter) {
        $parameters[] = $parameter->getName();
      }

      $row = [];
      $row['entity_type'] = $entity_type;
      $row['label'] = $attribute_instance->adminLabel;
      $row['method'] = $method_name;
      $row['method_plural'] = $method_plural;
      $row['parameters'] = [
        'data' => [
          '#theme' => 'item_list',
          '#items' => $parameters,
        ],
      ];

      $rows[$entity_type . '-' . $method_name] = $row;
    }
  }

}
